from typing import Union, Tuple

import numpy as np

from gui import CharactersLayer
from gui.opengl import OpenGLGUIFacade
from gui.opengl.OpenGLMesh import OpenGLMesh


class OpenGLCharactersLayer(OpenGLMesh, CharactersLayer):

    def __init__(self, gui: OpenGLGUIFacade):
        super(OpenGLCharactersLayer, self).__init__(gui)
        self.__vertices = None  # type: Union[None, np.ndarray]
        self.__faces = None  # type: Union[None, np.ndarray]
        self.__uvMap = None  # type: Union[None, np.ndarray]
        self.__transparentTile = (0, 0)  # type: Tuple[int, int]

    def setTileset(self, fileName: str, tileWidth: int, tileHeight: int):
        super(OpenGLCharactersLayer, self).setTileset(fileName, tileWidth, tileHeight)
        self.__transparentTile = (
            self.textureWidth // self.tileWidth - 1,
            self.textureHeight // self.tileHeight - 1
        )

    def draw(self):
        self.updateData(self.__vertices, self.__faces, self.__uvMap)
        super(OpenGLCharactersLayer, self).draw()

    def setCharacterCount(self, count: int):
        if self.__vertices is not None:
            raise RuntimeError("Update of characters count is not supported")

        self.__vertices = np.empty([count, 4, 2], dtype=np.float32)
        self.__uvMap = np.empty([count, 4, 2], dtype=np.float32)
        self.__faces = np.empty([count, 4], dtype=np.uint)
        faceCount = 0
        for charIndex in range(count):
            self.setCharacterLocation(charIndex, 0, 0)
            self.setCharacterTile(charIndex, self.__transparentTile[0], self.__transparentTile[1])

            self.__faces[charIndex, 0] = faceCount * 4
            self.__faces[charIndex, 1] = faceCount * 4 + 1
            self.__faces[charIndex, 2] = faceCount * 4 + 2
            self.__faces[charIndex, 3] = faceCount * 4 + 3
            faceCount += 1

        self.setData(self.__vertices, self.__faces, self.__uvMap, dynamic=True)

    def setCharacterLocation(self, charIndex: int, x: float, y: float):
        assert 0 <= charIndex < self.__vertices.shape[0]

        spriteScreenX1 = -1 + x * self.screenTileWidth
        spriteScreenY1 = 1 - y * self.screenTileHeight
        spriteScreenX2 = spriteScreenX1 + self.screenTileWidth
        spriteScreenY2 = spriteScreenY1 - self.screenTileHeight
        self.__vertices[charIndex, 0] = [spriteScreenX1, spriteScreenY2]
        self.__vertices[charIndex, 1] = [spriteScreenX1, spriteScreenY1]
        self.__vertices[charIndex, 2] = [spriteScreenX2, spriteScreenY1]
        self.__vertices[charIndex, 3] = [spriteScreenX2, spriteScreenY2]

    def setCharacterTile(self, charIndex: int, tileX: float, tileY: float):
        assert 0 <= charIndex < self.__vertices.shape[0]

        spriteTextureX1 = tileX * self.textureTileWidth
        spriteTextureY1 = tileY * self.textureTileHeight
        spriteTextureX2 = spriteTextureX1 + self.textureTileWidth
        spriteTextureY2 = spriteTextureY1 + self.textureTileHeight
        self.__uvMap[charIndex, 0] = [spriteTextureX1, spriteTextureY2]
        self.__uvMap[charIndex, 1] = [spriteTextureX1, spriteTextureY1]
        self.__uvMap[charIndex, 2] = [spriteTextureX2, spriteTextureY1]
        self.__uvMap[charIndex, 3] = [spriteTextureX2, spriteTextureY2]
