import logging
from typing import Union, Dict

import numpy as np
import pygame

from gui import TextLayer


class CharacterTile:
    __slots__ = ['x1', 'y1', 'x2', 'y2', 'advance']

    def __init__(self, x1: int = 0, y1: int = 0, x2: int = 0, y2: int = 0, advance: int = None):
        self.x1 = x1
        self.y1 = y1
        self.x2 = x2
        self.y2 = y2
        if advance is None:
            self.advance = x2 - x1
        else:
            self.advance = advance

    @property
    def width(self) -> int:
        return self.x2 - self.x1

    @property
    def height(self) -> int:
        return self.y2 - self.y1


class TextTileset:
    def __init__(self):
        self.__font = None  # type: Union[None, pygame.font.Font]
        self.__color = (255, 255, 255)  # type: (int, int, int)
        self.__color2 = (0, 0, 0)  # type: (int, int, int)
        self.__style = 0  # type: int
        self.__underlineShift = 0  # type: int
        self.__underlineThickness = 0  # type: int
        self.__shadowShift = 0  # type: int
        self.__shadowColor = (0, 0, 0)  # type: (int, int, int)
        self.__outlineThickness = 0  # type: int
        self.__outlineColor = (0, 0, 0)  # type: (int, int, int)
        self.__tileset = None  # type: Union[None, pygame.Surface]
        self.__nextTileX = 0  # type: int
        self.__nextTileY = 0  # type: int
        self.__tileDefaultWidth = 0  # type: int
        self.__tileDefaultHeight = 0  # type: int
        self.__characterTiles = {}  # type: Dict[str, CharacterTile]

    @property
    def tileDefaultWidth(self) -> int:
        return self.__tileDefaultWidth

    @property
    def tileDefaultHeight(self) -> int:
        return self.__tileDefaultHeight

    def __resize(self, width: int, height: int):

        # GPUs prefer square image size with a power of two
        def nextPowerOfTwo(x: int) -> int:
            return 1 if x == 0 else 2 ** (x - 1).bit_length()

        imageWidth = nextPowerOfTwo(width)
        imageHeight = nextPowerOfTwo(height)
        imageSize = max(imageWidth, imageHeight)
        self.__tileset = pygame.Surface((imageSize, imageSize), flags=pygame.SRCALPHA)

        # Copy current characters in the new tileset
        currentCharacters = "".join(self.__characterTiles.keys())
        self.__characterTiles.clear()
        self.__nextTileX = 0
        self.__nextTileY = 0
        self.addCharacters(currentCharacters)

    def __renderCharacter(self, character: str):
        surface = self.__font.render(character, False, self.__color)

        if self.__style & TextLayer.UNDERLINE:
            width = surface.get_width()
            start = (0, self.__underlineShift)
            end = (width - 1, self.__underlineShift)
            pygame.draw.line(surface, self.__color, start, end, self.__underlineThickness)

        if self.__style & TextLayer.SHADOW:
            # Create a new, larger surface
            shift = self.__shadowShift
            size = (surface.get_width() + shift, surface.get_height() + shift)
            shadowedSurface = pygame.Surface(size, flags=pygame.SRCALPHA)

            # Render the character with the shadow color
            surface2 = self.__font.render(character, False, self.__shadowColor)
            if self.__style & TextLayer.UNDERLINE:
                width = surface2.get_width()
                start = (0, self.__underlineShift)
                end = (width - 1, self.__underlineShift)
                pygame.draw.line(surface2, self.__shadowColor, start, end, self.__underlineThickness)

            # Blit the shadow and the character
            shadowedSurface.blit(surface2, (shift, shift))
            shadowedSurface.blit(surface, (0, 0))

            # Rendering with a shadow becomes the final surface
            surface = shadowedSurface
        elif self.__style & TextLayer.OUTLINE:
            # Create a new, larger surface
            thickness = self.__outlineThickness
            size = (surface.get_width() + 2 * thickness, surface.get_height() + 2 * thickness)
            outlineSurface = pygame.Surface(size, flags=pygame.SRCALPHA)

            # Blit the character
            outlineSurface.blit(surface, (thickness, thickness))
            # Draw the outline
            outlineColor = self.__outlineColor + (255, )
            for step in range(thickness):
                outlineSurface = outlineStep(outlineSurface, outlineColor)

            # Rendering with an outline becomes the final surface
            surface = outlineSurface

        return surface

    def setShadowColor(self, color: (int, int, int)):
        self.__shadowColor = color

    def setOutlineColor(self, color: (int, int, int)):
        self.__outlineColor = color

    def setFont(self, fontName: str, tileHeight: int, color: (int, int, int), style: int):
        # Font properties
        self.__color = color
        self.__style = style

        # If not set, automatically compute a underline properties
        if self.__underlineShift == 0:
            self.__underlineShift = (tileHeight * 90) // 100
        if self.__underlineThickness == 0:
            self.__underlineThickness = (tileHeight * 5) // 100
            if self.__underlineThickness < 1:
                self.__underlineThickness = 1

        # If not set, automatically compute a shift value for the shadow
        if self.__shadowShift == 0:
            shift = tileHeight // 16
            if shift < 1: shift = 1
            self.__shadowShift = shift

        # If not set, automatically compute a outline thickness
        if self.__outlineThickness == 0:
            thickness = tileHeight // 32
            if thickness < 1: thickness = 1
            self.__outlineThickness = thickness

        # Render a single character to find the required tile height
        fontSize = int(tileHeight * 1.3)
        for iterations in range(100):  # Up to 100 tries
            self.__font = pygame.font.Font(fontName, fontSize)
            if self.__style & TextLayer.ITALIC:
                self.__font.set_italic(True)
            if self.__style & TextLayer.BOLD:
                self.__font.set_bold(True)
            surface = self.__renderCharacter("M")
            if surface.get_height() == tileHeight:
                break
            elif surface.get_height() > tileHeight:
                fontSize -= 1
            else:
                fontSize += 1

        self.__tileDefaultWidth = surface.get_width()
        self.__tileDefaultHeight = surface.get_height()
        self.__characterTiles = {}
        self.__resize(256, 256)

    def addCharacters(self, characters: str) -> bool:
        tilesetUpdated = False
        for character in characters:
            if character in self.__characterTiles:
                continue
            if character == "\n":
                continue
            tilesetUpdated = True

            # Render the character
            surface = self.__renderCharacter(character)
            characterWidth = surface.get_width()
            if self.__tileDefaultHeight < surface.get_height():
                logging.warning("The height of character {} is higher than the default.".format(character))

            # Ensure that next tile location in inside the tileset
            if (self.__nextTileX + characterWidth) > self.__tileset.get_width():
                self.__nextTileX = 0
                self.__nextTileY += self.__tileDefaultHeight
            if (self.__nextTileY + self.__tileDefaultHeight) > self.__tileset.get_height():
                self.__resize(2 * self.__tileset.get_width(), 2 * self.__tileset.get_height())
                if (self.__nextTileX + characterWidth) > self.__tileset.get_width():
                    self.__nextTileX = 0
                    self.__nextTileY += self.__tileDefaultHeight

            # Blit the character in the tileset
            self.__tileset.blit(surface, (self.__nextTileX, self.__nextTileY))

            # Get metrics
            metrics = self.__font.metrics(character)
            advance = metrics[0][4]
            if self.__style & TextLayer.OUTLINE:
                advance += self.__outlineThickness * 2

            # Compute and save the tile location
            self.__characterTiles[character] = CharacterTile(
                self.__nextTileX, self.__nextTileY,
                self.__nextTileX + characterWidth, self.__nextTileY + self.__tileDefaultHeight,
                advance
            )

            # Update tile next location
            self.__nextTileX += characterWidth
        return tilesetUpdated

    def toNumpyArray(self) -> np.ndarray:
        imageArray = np.zeros((self.__tileset.get_width(), self.__tileset.get_height(), 4), dtype=np.uint8)
        imageArray[..., 0] = pygame.surfarray.pixels_red(self.__tileset)
        imageArray[..., 1] = pygame.surfarray.pixels_green(self.__tileset)
        imageArray[..., 2] = pygame.surfarray.pixels_blue(self.__tileset)
        imageArray[..., 3] = pygame.surfarray.pixels_alpha(self.__tileset)
        return imageArray.transpose((1, 0, 2))

    def getCharacterTile(self, character: str) -> CharacterTile:
        return self.__characterTiles[character]


def outlineStep(surface: pygame.Surface, outlineColor: (int, int, int, int)) -> pygame.Surface:
    result = surface.copy()
    w = surface.get_width()
    h = surface.get_height()
    for y in range(h):
        for x in range(w):
            # Copy opaque pixels
            color = surface.get_at((x, y))
            if color[3] != 0:
                result.set_at((x, y), color)
                continue
            # If any pixel around the current one is opaque, then light it
            if (x > 0 and surface.get_at((x - 1, y))[3] != 0) \
                    or (x < w - 1 and surface.get_at((x + 1, y))[3] != 0) \
                    or (y > 0 and surface.get_at((x, y - 1))[3] != 0) \
                    or (y < h - 1 and surface.get_at((x, y + 1))[3] != 0) \
                    or (x > 0 and y > 0 and surface.get_at((x - 1, y - 1))[3] != 0) \
                    or (x > 0 and y < h - 1 and surface.get_at((x - 1, y + 1))[3] != 0) \
                    or (x < w - 1 and y > 0 and surface.get_at((x + 1, y - 1))[3] != 0) \
                    or (x < w - 1 and y < h - 1 and surface.get_at((x + 1, y + 1))[3] != 0):
                result.set_at((x, y), outlineColor)
    return result
